from __future__ import absolute_import

import typing
from datetime import datetime

from applitools.common import FloatingMatchSettings, MatchResult, RunningSession, logger
from applitools.common.errors import OutOfBoundsError
from applitools.common.geometry import Point, Region, AccessibilityRegion
from applitools.common.match import ImageMatchSettings
from applitools.common.match_window_data import MatchWindowData, Options
from applitools.common.utils import datetime_utils, image_utils
from applitools.common.ultrafastgrid import VisualGridSelector
from applitools.core.capture import AppOutputProvider, AppOutputWithScreenshot

from .fluent import CheckSettings, GetFloatingRegion, GetRegion, GetAccessibilityRegion

if typing.TYPE_CHECKING:
    from typing import List, Text, Optional, Union
    from applitools.common import EyesScreenshot
    from applitools.common.utils.custom_types import Num, UserInputs
    from applitools.core.server_connector import ServerConnector
    from .eyes_base import EyesBase

    GET_REGION = Union[GetRegion, GetFloatingRegion]
    REGION = Union[Region, FloatingMatchSettings]

__all__ = ("MatchWindowTask",)


def _collect_regions(region_providers, screenshot, eyes):
    # type: (List[GET_REGION], EyesScreenshot, EyesBase) -> List[REGION]
    collected = []  # type: List[REGION]
    for provider in region_providers:
        try:
            regions = provider.get_regions(eyes, screenshot)
            collected.extend(regions)
        except OutOfBoundsError:
            logger.warning("Region was out of bounds")
    return collected


def filter_empty_entries(regions, location):
    return [region.offset(-location) for region in regions if region.area != 0]


def filter_empty_entries_and_combine(regions, location, region_selectors):
    for i, reg in enumerate(regions):
        if reg.area == 0:
            continue
        reg = reg.offset(-location)
        vgs = region_selectors[i]
        yield reg, vgs.category


def collect_regions_from_selectors(image_match_settings, regions, region_selectors):
    # type:(ImageMatchSettings,List[Region],List[List[VisualGridSelector]])->ImageMatchSettings
    if not regions:
        return image_match_settings
    logger.debug(
        """
    start collect_regions_from_selectors()

        regions: {}
        region_selectors: {}
    """.format(
            regions, region_selectors
        )
    )
    mutable_regions = [
        [],  # Ignore Regions
        [],  # Layout Regions
        [],  # Strict Regions
        [],  # Content Regions
        [],  # Floating Regions
        [],  # Accessibility Regions
        [],  # Target Element Location
    ]
    r_selector_counts = [len(r) for r in region_selectors]  # mapping of
    prev_count = 0
    for selectors_count, m_specific_regions in zip(r_selector_counts, mutable_regions):
        if selectors_count == 0:
            continue
        next_count = prev_count + selectors_count
        m_specific_regions.extend(regions[prev_count:next_count])
        prev_count = next_count

    location = Point.ZERO()

    # If target element location available
    selector_regions_index = len(mutable_regions) - 1
    if mutable_regions[selector_regions_index]:
        location = mutable_regions[selector_regions_index][0].location

    image_match_settings.ignore_regions = filter_empty_entries(
        mutable_regions[0], location
    )
    image_match_settings.layout_regions = filter_empty_entries(
        mutable_regions[1], location
    )
    image_match_settings.strict_regions = filter_empty_entries(
        mutable_regions[2], location
    )
    image_match_settings.content_regions = filter_empty_entries(
        mutable_regions[3], location
    )
    image_match_settings.floating_match_settings = [
        FloatingMatchSettings(reg, gfr.floating_bounds)
        for (reg, gfr) in filter_empty_entries_and_combine(
            mutable_regions[4], location, region_selectors[4]
        )
        if isinstance(gfr, GetFloatingRegion)
    ]
    image_match_settings.accessibility = [
        AccessibilityRegion.from_(reg, gfr.accessibility_type)
        for (reg, gfr) in filter_empty_entries_and_combine(
            mutable_regions[5], location, region_selectors[5]
        )
        if isinstance(gfr, GetAccessibilityRegion)
    ]
    logger.debug(
        """
    finish collect_regions_from_selectors()

        image_match_settings: {}
    """.format(
            image_match_settings
        )
    )
    return image_match_settings


def collect_regions_from_screenshot(
    check_settings, image_match_settings, screenshot, eyes
):
    # type:(CheckSettings,ImageMatchSettings,EyesScreenshot,EyesBase)->ImageMatchSettings

    image_match_settings.ignore_regions = _collect_regions(  # type: ignore
        check_settings.values.ignore_regions, screenshot, eyes
    )
    image_match_settings.layout_regions = _collect_regions(  # type: ignore
        check_settings.values.layout_regions, screenshot, eyes
    )
    image_match_settings.strict_regions = _collect_regions(  # type: ignore
        check_settings.values.strict_regions, screenshot, eyes
    )
    image_match_settings.content_regions = _collect_regions(  # type: ignore
        check_settings.values.content_regions, screenshot, eyes
    )
    image_match_settings.floating_match_settings = _collect_regions(  # type: ignore
        check_settings.values.floating_regions, screenshot, eyes
    )
    image_match_settings.accessibility = _collect_regions(  # type: ignore
        check_settings.values.accessibility_regions, screenshot, eyes
    )
    logger.debug(
        """
    finish collect_regions_from_screenshot()

        image_match_settings: {}
    """.format(
            image_match_settings
        )
    )
    return image_match_settings


class MatchWindowTask(object):
    """
    Handles matching of output with the expected output
    (including retry and 'ignore mismatch' when needed).
    """

    _MATCH_INTERVAL_MS = 500  # ms

    def __init__(
        self,
        server_connector,  # type: ServerConnector
        running_session,  # type: RunningSession
        retry_timeout_ms,  # type: Num
        eyes,  # type: EyesBase
        app_output_provider=None,  # type: Optional[AppOutputProvider]
    ):
        # type: (...) -> None
        """
        Ctor.

        :param server_connector: The agent connector to use for communication.
        :param running_session:  The current eyes session.
        :param retry_timeout_ms: The default match timeout. (milliseconds)
        :param eyes: The Eyes instance which created this task.
        :param app_output_provider: A callback for getting the application output
                                    when performing match.
       """
        self._server_connector = server_connector
        self._running_session = running_session
        self._eyes = eyes
        self._app_output_provider = app_output_provider
        self._default_retry_timeout_ms = retry_timeout_ms  # type: Num

        self._match_result = None  # type: Optional[MatchResult]
        self._last_screenshot = None  # type: Optional[EyesScreenshot]

    @staticmethod
    def create_image_match_settings(check_settings, eyes, screenshot=None):
        # type: (CheckSettings, EyesBase, Options[EyesScreenshot])-> ImageMatchSettings
        img = ImageMatchSettings.create_from(eyes.configure.default_match_settings)
        img.accessibility_settings = eyes.configure.accessibility_validation

        # Set defaults if necessary
        if check_settings.values.match_level is not None:
            img.match_level = check_settings.values.match_level
        if check_settings.values.ignore_caret is not None:
            img.ignore_caret = check_settings.values.ignore_caret
        if check_settings.values.use_dom is not None:
            img.use_dom = check_settings.values.use_dom
        if check_settings.values.enable_patterns is not None:
            img.enable_patterns = check_settings.values.enable_patterns
        if check_settings.values.ignore_displacements is not None:
            img.ignore_displacements = check_settings.values.ignore_displacements
        if check_settings.values.match_level is not None:
            img.match_level = check_settings.values.match_level

        if screenshot:
            img = collect_regions_from_screenshot(check_settings, img, screenshot, eyes)
        return img

    @property
    def last_screenshot(self):
        # type: () -> EyesScreenshot
        return self._last_screenshot

    @property
    def last_screenshot_bounds(self):
        # type: () -> Region
        return self._last_screenshot_bounds

    def match_window(
        self,
        user_inputs,  # type: UserInputs
        region,  # type: Region
        tag,  # type: Text
        should_run_once_on_timeout,  # type: bool
        ignore_mismatch,  # type: bool
        check_settings,  # type: CheckSettings
        retry_timeout_ms,  # type: Num
    ):
        # type: (...) -> MatchResult
        if retry_timeout_ms is None or retry_timeout_ms < 0:
            retry_timeout_ms = self._default_retry_timeout_ms
        logger.debug("retry_timeout = {} ms".format(retry_timeout_ms))

        screenshot = self._take_screenshot(
            user_inputs,
            region,
            tag,
            should_run_once_on_timeout,
            ignore_mismatch,
            check_settings,
            retry_timeout_ms,
        )
        if ignore_mismatch:
            return self._match_result
        self._update_last_screenshot(screenshot)
        self._update_bounds(region)
        return self._match_result

    def perform_match(
        self,
        app_output,  # type: AppOutputWithScreenshot
        name,  # type: Text
        ignore_mismatch,  # type: bool
        image_match_settings,  # type: ImageMatchSettings
        eyes,  # type: EyesBase
        user_inputs=None,  # type: Optional[UserInputs]
        regions=None,  # type: Optional[List[Region]]
        region_selectors=None,  # type: Optional[List[List[VisualGridSelector]]]
        check_settings=None,  # type: Optional[CheckSettings]
        render_id=None,  # type: Optional[Text]
    ):
        # type: (...) -> MatchResult

        screenshot = app_output.screenshot
        if check_settings and screenshot:
            image_match_settings = collect_regions_from_screenshot(
                check_settings, image_match_settings, screenshot, eyes
            )
        elif regions and region_selectors:
            # visual grid
            image_match_settings = collect_regions_from_selectors(
                image_match_settings, regions, region_selectors
            )

        user_inputs = user_inputs or []
        agent_setup = self._eyes.agent_setup
        return self._perform_match(
            user_inputs,
            app_output,
            name,
            ignore_mismatch,
            image_match_settings,
            agent_setup,
            render_id,
        )

    def _perform_match(
        self,
        user_inputs,  # type: UserInputs
        app_output_width_screenshot,  # type: AppOutputWithScreenshot
        name,  # type: Text
        ignore_mismatch,  # type: bool
        image_match_settings,  # type: ImageMatchSettings
        agent_setup,  # type: Text
        render_id,  # type: Text
    ):
        # type: (...) -> MatchResult

        screenshot = app_output_width_screenshot.screenshot
        app_output = app_output_width_screenshot.app_output
        if screenshot:
            app_output.screenshot_bytes = image_utils.get_bytes(screenshot.image)

        match_window_data = MatchWindowData(
            ignore_mismatch=ignore_mismatch,
            user_inputs=user_inputs,
            options=Options(
                name=name,
                user_inputs=user_inputs,
                ignore_mismatch=ignore_mismatch,
                ignore_match=False,
                force_mismatch=False,
                force_match=False,
                image_match_settings=image_match_settings,
                render_id=render_id,
            ),
            app_output=app_output,
            tag=name,
            agent_setup=agent_setup,
            render_id=render_id,
        )
        return self._server_connector.match_window(
            self._running_session, match_window_data
        )

    def _update_last_screenshot(self, screenshot):
        # type: (EyesScreenshot) -> None
        if screenshot:
            self._last_screenshot = screenshot

    def _update_bounds(self, region):
        # type: (Region) -> None
        if region.is_size_empty:
            if self._last_screenshot:
                self._last_screenshot_bounds = Region(
                    0,
                    0,
                    self._last_screenshot.image.width,
                    self._last_screenshot.image.height,
                )
            else:
                # We set an "infinite" image size since we don't know what the
                # screenshot size is...
                self._last_screenshot_bounds = Region(0, 0, 999999999999, 999999999999)
        else:
            self._last_screenshot_bounds = region

    def _take_screenshot(
        self,
        user_inputs,  # type: UserInputs
        region,  # type: Region
        tag,  # type: Text
        should_run_once_on_timeout,  # type: bool
        ignore_mismatch,  # type: bool
        check_settings,  # type: CheckSettings
        retry_timeout,  # type: int
    ):
        # type: (...) -> EyesScreenshot

        time_start = datetime.now()
        if retry_timeout == 0 or should_run_once_on_timeout:
            if should_run_once_on_timeout:
                datetime_utils.sleep(retry_timeout)

            screenshot = self._try_take_screenshot(
                user_inputs, region, tag, ignore_mismatch, check_settings
            )
        else:
            screenshot = self._retry_taking_screenshot(
                user_inputs, region, tag, ignore_mismatch, check_settings, retry_timeout
            )
        time_end = datetime.now()
        summary_ms = datetime_utils.to_ms((time_end - time_start).seconds)
        logger.debug(
            "MatchWindowTask._take_screenshot completed in {} ms".format(summary_ms)
        )
        return screenshot

    def _try_take_screenshot(
        self,
        user_inputs,  # type: UserInputs
        region,  # type: Region
        tag,  # type: Text
        ignore_mismatch,  # type: bool
        check_settings,  # type: CheckSettings
    ):
        # type: (...) -> EyesScreenshot
        app_output = self._app_output_provider.get_app_output(
            region, self._last_screenshot, check_settings
        )
        image_match_settings = self.create_image_match_settings(
            check_settings, self._eyes
        )
        self._match_result = self.perform_match(
            app_output,
            tag,
            ignore_mismatch,
            image_match_settings,
            self._eyes,
            user_inputs,
            check_settings=check_settings,
        )
        return app_output.screenshot

    def _retry_taking_screenshot(
        self,
        user_inputs,  # type: UserInputs
        region,  # type: Region
        tag,  # type: Text
        ignore_mismatch,  # type: bool
        check_settings,  # type: CheckSettings
        retry_timeout_sec,  # type: int
    ):
        # type: (...) -> EyesScreenshot

        start = datetime.now()  # Start the retry timer.
        retry = datetime_utils.to_ms((datetime.now() - start).seconds)

        # The match retry loop.
        screenshot = self._taking_screenshot_loop(
            user_inputs,
            region,
            tag,
            ignore_mismatch,
            check_settings,
            retry_timeout_sec,
            retry,
            start,
        )
        # If we're here because we haven't found a match yet, try once more
        if not self._match_result.as_expected:
            logger.info("Window mismatch. Retrying...")
            return self._try_take_screenshot(
                user_inputs, region, tag, ignore_mismatch, check_settings
            )
        return screenshot

    def _taking_screenshot_loop(
        self,
        user_inputs,  # type: UserInputs
        region,  # type: Region
        tag,  # type: Text
        ignore_mismatch,  # type: bool
        check_settings,  # type: CheckSettings
        retry_timeout_ms,  # type: int
        retry_ms,  # type: int
        start,  # type: datetime
        screenshot=None,  # type: Optional[EyesScreenshot]
    ):
        # type: (...) -> Optional[EyesScreenshot]

        if retry_ms >= retry_timeout_ms:
            return screenshot

        datetime_utils.sleep(self._MATCH_INTERVAL_MS)

        new_screenshot = self._try_take_screenshot(
            user_inputs,
            region,
            tag,
            ignore_mismatch=True,
            check_settings=check_settings,
        )
        if self._match_result.as_expected:
            return new_screenshot

        retry_ms = (start - datetime.now()).seconds
        return self._taking_screenshot_loop(
            user_inputs,
            region,
            tag,
            ignore_mismatch,
            check_settings,
            retry_timeout_ms,
            retry_ms,
            start,
            new_screenshot,
        )



from __future__ import absolute_import

from typing import TYPE_CHECKING

from applitools.common import ScaleProvider, logger

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import ViewPort

__all__ = ("FixedScaleProvider", "NullScaleProvider", "ContextBasedScaleProvider")


class FixedScaleProvider(ScaleProvider):
    def __init__(self, scale_ratio, *args, **kwargs):
        super(FixedScaleProvider, self).__init__(scale_ratio)
        self._scale_ratio = scale_ratio


class NullScaleProvider(FixedScaleProvider):
    UNKNOWN_SCALE_RATIO = 1.0

    def __init__(self, *args, **kwargs):
        super(NullScaleProvider, self).__init__(self.UNKNOWN_SCALE_RATIO)


class ContextBasedScaleProvider(ScaleProvider):
    _ALLOWED_VS_DEVIATION = 1
    _ALLOWED_DCES_DEVIATION = 10

    def __init__(
        self,
        top_level_context_entire_size,  # type: ViewPort
        viewport_size,  # type: ViewPort
        device_pixel_ratio,  # type: float
        is_mobile_device,  # type: bool
    ):
        super(ContextBasedScaleProvider, self).__init__()
        self.top_level_context_entire_size = top_level_context_entire_size
        self.viewport_size = viewport_size
        self.device_pixel_ratio = device_pixel_ratio
        self.is_mobile_device = is_mobile_device

    @property
    def scale_ratio(self):
        return self._scale_ratio

    @staticmethod
    def get_scale_ratio_to_viewport(
        viewport_width,  # type: float
        image_to_scale_width,  # type: float
        current_scale_ratio,  # type: float
    ):
        # type: (...) -> float
        scaled_image_width = round(image_to_scale_width * viewport_width)
        from_scaled_to_viewport_ratio = viewport_width / scaled_image_width
        return current_scale_ratio * from_scaled_to_viewport_ratio

    def update_scale_ratio(self, image_to_scale_width):
        # type: (float) -> None
        viewport_width = self.viewport_size["width"]
        dces_width = self.top_level_context_entire_size["width"]

        # If the image's width is the same as the viewport's width or the
        # top level context's width, no scaling is necessary.
        allowed_vs_deviation = (
            viewport_width - self._ALLOWED_VS_DEVIATION
            <= image_to_scale_width
            <= viewport_width + self._ALLOWED_VS_DEVIATION
        )
        allowed_dces_deviation = (
            dces_width - self._ALLOWED_DCES_DEVIATION
            <= image_to_scale_width
            <= dces_width + self._ALLOWED_DCES_DEVIATION
        )
        if allowed_dces_deviation or allowed_vs_deviation:
            logger.info("Image is already scaled correctly.")
            self._scale_ratio = 1.0
        else:
            logger.info("Calculating the scale ratio..")
            self._scale_ratio = 1.0 / self.device_pixel_ratio
            if self.is_mobile_device:
                logger.info(
                    "Mobile device, so using 2 step calculation for scale ration..."
                )
                logger.info("Scale ratio based on DRP: {}".format(self._scale_ratio))
                self._scale_ratio = self.get_scale_ratio_to_viewport(
                    viewport_width, image_to_scale_width, self._scale_ratio
                )
            logger.info("Final scale ratio: {}".format(self.scale_ratio))



import os
from abc import abstractmethod
from datetime import datetime

import attr

from applitools.common.utils import image_utils
from applitools.common.utils.general_utils import get_env_with_prefix


@attr.s
class DebugScreenshotProvider(object):
    """Interface for saving debug screenshots."""

    DEFAULT_PREFIX = get_env_with_prefix("DEBUG_SCREENSHOT_PREFIX", "screenshot_")
    DEFAULT_PATH = get_env_with_prefix("DEBUG_SCREENSHOT_PATH", "")

    _prefix = attr.ib(default=DEFAULT_PREFIX)
    _path = attr.ib(default=DEFAULT_PATH)

    def __attrs_post_init__(self):
        self._image_counter = 0

    @property
    def prefix(self):
        return self._prefix

    @prefix.setter
    def prefix(self, value):
        if value:
            self._prefix = value

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        if value:
            self._path = value.rstrip("/")

    @abstractmethod
    def save(self, image, suffix):
        pass


class NullDebugScreenshotProvider(DebugScreenshotProvider):
    """A mock debug screenshot provider."""

    def save(self, image, suffix):
        pass


class FileDebugScreenshotProvider(DebugScreenshotProvider):
    """ A debug screenshot provider for saving screenshots to file."""

    def save(self, image, suffix):
        now = datetime.now().strftime("%H:%M:%S")
        suffix = "{}-{}".format(self._image_counter, suffix)
        self._image_counter += 1

        filename = "{path}/{prefix}_{timestamp}_{suffix}.png".format(
            path=self.path, prefix=self.DEFAULT_PREFIX, timestamp=now, suffix=suffix
        )
        image_utils.save_image(image, filename)



from applitools.common.geometry import Region
from applitools.common.utils import ABC
from applitools.core.fluent.check_settings import CheckSettings

_all_ = ("CheckTarget",)


class CheckTarget(ABC):
    @staticmethod
    def window():
        # type: () -> CheckSettings
        pass

    @staticmethod
    def region(region):
        # type: (Region) -> CheckSettings
        pass



import math
from abc import abstractmethod
from typing import Dict, Union

import attr
from PIL.Image import Image

from applitools.common import RectangleSize, Region
from applitools.common.utils import ABC, image_utils

__all__ = ("FixedCutProvider", "UnscaledFixedCutProvider", "NullCutProvider")


@attr.s
class CutProvider(ABC):
    header = attr.ib()  # type: int
    footer = attr.ib()  # type: int
    left = attr.ib()  # type: int
    right = attr.ib()  # type: int

    def cut(self, image):
        # type: (Image) -> Image
        if self.header == 0 and self.footer == 0 and self.right == 0 and self.left == 0:
            return image

        target_region = Region(
            self.left,
            self.header,
            image.width - self.left - self.right,
            image.height - self.header - self.footer,
        )
        return image_utils.crop_image(image, target_region)

    @abstractmethod
    def scale(self, scale_ratio):
        # type: (float) -> CutProvider
        """Get a scaled version of the cut provider"""

    def to_region(self, size):
        # type: (Union[RectangleSize, Dict]) -> Region
        return Region(
            left=self.left,
            top=self.header,
            width=size["width"] - self.left - self.right,
            height=size["height"] - self.header - self.footer,
        )


class FixedCutProvider(CutProvider):
    def scale(self, scale_ratio):
        # type: (float) -> FixedCutProvider
        cut_provider = FixedCutProvider(
            math.ceil(self.header * scale_ratio),
            math.ceil(self.footer * scale_ratio),
            math.ceil(self.left * scale_ratio),
            math.ceil(self.right * scale_ratio),
        )
        return cut_provider


class UnscaledFixedCutProvider(CutProvider):
    def scale(self, scale_ratio):
        # type: (float) -> UnscaledFixedCutProvider
        if isinstance(self, NullCutProvider):
            return self

        cut_provider = UnscaledFixedCutProvider(
            self.header, self.footer, self.left, self.right
        )
        return cut_provider


class NullCutProvider(UnscaledFixedCutProvider):
    def __init__(self):
        super(NullCutProvider, self).__init__(0, 0, 0, 0)



from .check_settings import CheckSettings, CheckSettingsValues
from .check_target import CheckTarget
from .region import (
    FloatingRegionByRectangle,
    GetFloatingRegion,
    GetRegion,
    RegionByRectangle,
    GetAccessibilityRegion,
    AccessibilityRegionByRectangle,
)

__all__ = (
    "CheckSettings",
    "CheckSettingsValues",
    "CheckTarget",
    "GetRegion",
    "GetFloatingRegion",
    "FloatingRegionByRectangle",
    "RegionByRectangle",
    "GetAccessibilityRegion",
    "AccessibilityRegionByRectangle",
)



from applitools.common import logger, Configuration


class EyesConfigurationMixin(object):
    _config_cls = Configuration

    def __init__(self):
        self._config_provider = self._config_cls()

    def get_configuration(self):
        # type:() -> Configuration
        """Returns clone of configuration instance"""
        if isinstance(self._config_provider, self._config_cls):
            return self._config_provider.clone()
        return self._config_provider.configure.clone()

    def set_configuration(self, configuration):
        # type:(Configuration) -> None
        """Clone configuration instance and set it"""
        old_configuration = self._config_provider
        if isinstance(old_configuration, self._config_cls):
            new_configuration = configuration.clone()
            if old_configuration.api_key and not configuration.api_key:
                new_configuration.api_key = old_configuration.api_key
            if old_configuration.server_url and not configuration.server_url:
                new_configuration.server_url = old_configuration.server_url
            self._config_provider = new_configuration
        else:
            self._config_provider = configuration

    @property
    def configure(self):
        # type:() -> Configuration
        if isinstance(self._config_provider, self._config_cls):
            return self._config_provider
        return self._config_provider.configure

    @property
    def configuration(self):
        logger.deprecation("`configuration` is deprecated. Use `configure` instead")
        return self.configure

    @configuration.setter
    def configuration(self, configuration):
        logger.deprecation(
            "Assign to `configuration` is deprecated. "
            "Use `set_configuration` instead"
        )
        self.set_configuration(configuration)



import abc
import typing

from applitools.common.geometry import Point
from applitools.common.utils import ABC, iteritems

if typing.TYPE_CHECKING:
    from typing import List, Optional, Any
    from applitools.common import RectangleSize


class PositionMemento(object):
    """
    A base class for position related memento instances. This is intentionally
    not an interface, since the mementos might vary in their interfaces.
    """

    def __init__(self, position, **kwargs):
        # type: (Point, **Any) -> None
        self.position = position
        for name, val in iteritems(kwargs):
            setattr(self, name, val)

    def __str__(self):
        return "PositionMemento(position={}, ...)".format(self.position)


class PositionProvider(ABC):
    @abc.abstractmethod
    def get_current_position(self):
        # type: () -> Optional[Point]
        """
        :return: The current position, or `None` if position is not available.
        """

    @abc.abstractmethod
    def set_position(self, location):
        # type: (Point) -> Point
        """
        Go to the specified location.

        :param location: The position to set
        """

    @abc.abstractmethod
    def get_entire_size(self):
        # type: () -> RectangleSize
        """
        :return: The entire size of the container which the position is relative to.
        """

    def get_state(self):
        # type: () -> PositionMemento
        """Get the current state of the position provider.

        This is different from :eyes_selenium_utils:`get_current_position()` in that the
        state of the position provider
        might include other model than just the coordinates. For example a CSS
        translation based position provider (in WebDriver based SDKs), might
        save the entire "transform" style value as its state.

        Returns:
            The current state of the position provider, which can later be
            restored by  passing it as a parameter to :self:`restore_state`.
        """
        return PositionMemento(self.get_current_position())

    def restore_state(self, state):
        # type: (PositionMemento) -> None
        """ Restores the state of the position provider to the state provided as a
        parameter.

        Args:
            state: The state to restore to.
        """
        self.set_position(state.position)


class InvalidPositionProvider(PositionProvider):
    """
    An implementation of :py:class:`PositionProvider` which throws an exception
    for every method. Can be used as a placeholder until an actual
    implementation is set.
    """

    def get_current_position(self):
        raise NotImplementedError("This class does not implement methods!")

    def set_position(self, location):
        # type: (Point) -> None
        raise NotImplementedError("This class does not implement methods!")

    def get_entire_size(self):
        raise NotImplementedError("This class does not implement methods!")



from .position_provider import (  # noqa: F401
    InvalidPositionProvider,
    PositionMemento,
    PositionProvider,
)
from .region_provider import (  # noqa: F401
    NULL_REGION_PROVIDER,
    NullRegionProvider,
    RegionProvider,
)



import typing
from abc import abstractmethod

import attr

from applitools.common import AppOutput, EyesScreenshot, Point, Region
from applitools.common.utils import ABC

if typing.TYPE_CHECKING:
    from PIL.Image import Image
    from typing import Optional, Callable
    from applitools.core import CheckSettings, EyesBase

    T = typing.TypeVar("T", bound=CheckSettings)

__all__ = (
    "AppOutputWithScreenshot",
    "AppOutputProvider",
    "EyesScreenshotFactory",
    "ImageProvider",
)


@attr.s
class AppOutputWithScreenshot(object):
    app_output = attr.ib(type=AppOutput)  # type: AppOutput
    screenshot = attr.ib()  # type: Optional[EyesScreenshot]
    location = attr.ib(default=None)  # type: Optional[Point]


@attr.s
class AppOutputProvider(object):
    method = attr.ib()  # type: Callable

    def get_app_output(self, region, last_screenshot, check_settings):
        # type: (Region, EyesScreenshot, T) -> AppOutputWithScreenshot
        return self.method(region, last_screenshot, check_settings)


class EyesScreenshotFactory(ABC):
    """
    Encapsulates the instantiation of an EyesScreenshot object.
    """

    @abstractmethod
    def make_screenshot(self, image):
        # type: (Image) -> EyesScreenshot
        pass


@attr.s
class ImageProvider(ABC):
    """
    Encapsulates image retrieval.
    """

    _eyes = attr.ib()  # type: EyesBase

    @abstractmethod
    def get_image(self):
        # type: () -> Image
        pass



from .capture import (
    AppOutputProvider,
    AppOutputWithScreenshot,
    EyesScreenshotFactory,
    ImageProvider,
)
from .cut import FixedCutProvider, NullCutProvider, UnscaledFixedCutProvider
from .eyes_base import EyesBase
from .eyes_runner import EyesRunner
from .fluent import (
    CheckSettings,
    CheckSettingsValues,
    CheckTarget,
    FloatingRegionByRectangle,
    GetFloatingRegion,
    GetRegion,
    RegionByRectangle,
)
from .match_window_task import MatchWindowTask
from .positioning import (
    NULL_REGION_PROVIDER,
    InvalidPositionProvider,
    NullRegionProvider,
    PositionMemento,
    PositionProvider,
    RegionProvider,
)
from .scaling import (
    ContextBasedScaleProvider,
    FixedScaleProvider,
    NullScaleProvider,
    ScaleProvider,
)
from .server_connector import ServerConnector
from .triggers import MouseTrigger, TextTrigger
from .batch_close import BatchClose
from .locators import VisualLocator

__all__ = (
    "TextTrigger",
    "MouseTrigger",
    "MatchWindowTask",
    "ContextBasedScaleProvider",
    "FixedScaleProvider",
    "NullScaleProvider",
    "ScaleProvider",
    "EyesBase",
    "PositionProvider",
    "PositionMemento",
    "InvalidPositionProvider",
    "RegionProvider",
    "FixedCutProvider",
    "NullCutProvider",
    "UnscaledFixedCutProvider",
    "NullRegionProvider",
    "NULL_REGION_PROVIDER",
    "CheckSettings",
    "CheckSettingsValues",
    "CheckTarget",
    "GetRegion",
    "GetFloatingRegion",
    "FloatingRegionByRectangle",
    "RegionByRectangle",
    "ServerConnector",
    "AppOutputWithScreenshot",
    "AppOutputProvider",
    "EyesScreenshotFactory",
    "ImageProvider",
    "EyesRunner",
    "BatchClose",
    "VisualLocator",
)



from __future__ import absolute_import

import abc
import typing

from applitools.common import AppOutput, RectangleSize, Region, RunningSession, logger
from applitools.common.config import Configuration
from applitools.common.errors import (
    DiffsFoundError,
    EyesError,
    NewTestError,
    TestFailedError,
)
from applitools.common.match import MatchResult
from applitools.common.metadata import AppEnvironment, SessionStartInfo
from applitools.common.server import FailureReports, SessionType
from applitools.common.test_results import TestResults
from applitools.common.utils import ABC, argument_guard
from applitools.common.utils.compat import raise_from
from applitools.common.ultrafastgrid import RenderingInfo
from applitools.core.capture import AppOutputProvider, AppOutputWithScreenshot
from applitools.core.cut import (
    FixedCutProvider,
    NullCutProvider,
    UnscaledFixedCutProvider,
)
from applitools.core.debug import (
    FileDebugScreenshotProvider,
    NullDebugScreenshotProvider,
)
from applitools.core.eyes_mixins import EyesConfigurationMixin

from .match_window_task import MatchWindowTask
from .positioning import InvalidPositionProvider, PositionProvider, RegionProvider
from .scaling import FixedScaleProvider, NullScaleProvider, ScaleProvider
from .server_connector import ServerConnector

if typing.TYPE_CHECKING:
    from applitools.common.utils.custom_types import ViewPort, UserInputs, Num
    from applitools.core.fluent.check_settings import CheckSettings
    from applitools.common.capture import EyesScreenshot
    from applitools.common import MatchLevel
    from typing import Optional, Text, Union

__all__ = ("EyesBase",)


class _EyesBaseAbstract(ABC):
    @abc.abstractmethod
    def _try_capture_dom(self):
        # type: () -> Text
        """
        Returns the string with DOM of the current page in the prepared format or empty string
        """

    @property
    @abc.abstractmethod
    def base_agent_id(self):
        # type: () -> Text
        """
        Must return version of SDK. (e.g. Selenium, Images) in next format:
            "eyes.{package}.python/{lib_version}"
        """

    @abc.abstractmethod
    def _get_screenshot(self):
        # type: (...) -> EyesScreenshot
        pass

    @property
    @abc.abstractmethod
    def _title(self):
        # type: () -> Text
        """
        Returns the title of the window of the AUT, or empty string
         if the title is not available.
        """

    @abc.abstractmethod
    def _get_viewport_size(self):
        # type: () -> RectangleSize
        """

        :return:
        """

    @abc.abstractmethod
    def _set_viewport_size(self, size):
        # type: (ViewPort) -> None
        """

        """

    @property
    @abc.abstractmethod
    def _inferred_environment(self):
        pass


class EyesBase(EyesConfigurationMixin, _EyesBaseAbstract, ABC):
    _MAX_ITERATIONS = 10
    _running_session = None  # type: Optional[RunningSession]
    _session_start_info = None  # type: Optional[SessionStartInfo]
    _last_screenshot = None  # type: Optional[EyesScreenshot]
    _scale_provider = None  # type: Optional[ScaleProvider]
    _dom_url = None  # type: Optional[Text]
    _position_provider = None  # type: Optional[PositionProvider]
    _is_viewport_size_set = False  # type: bool
    _should_match_once_on_timeout = False  # type: bool
    _is_opened = False  # type: bool
    _render_info = None  # type: Optional[RenderingInfo]
    _render = False
    _cut_provider = None
    _should_get_title = False  # type: bool
    _config_cls = Configuration
    # TODO: make it run with no effect to other pices of code
    # def set_explicit_viewport_size(self, size):
    #     """
    #     Define the viewport size as {@code size} without doing any actual action on the
    #
    #     :param size: The size of the viewport.
    #     """
    #     if not size:
    #         self.viewport_size = None
    #         self._is_viewport_size_set = False
    #         return None
    #     logger.info("Viewport size explicitly set to {}".format(size))
    #     self.viewport_size = RectangleSize(size["width"], size["height"])
    #     self._is_viewport_size_set = True

    def __init__(self):
        """
        Creates a new (possibly disabled) Eyes instance that
        interacts with the Eyes server.
        """
        super(EyesBase, self).__init__()
        self._server_connector = ServerConnector()  # type: ServerConnector
        self._user_inputs = []  # type: UserInputs
        self._debug_screenshot_provider = NullDebugScreenshotProvider()

    @property
    def match_level(self):
        # type: () -> MatchLevel
        return self.configure.match_level

    @match_level.setter
    def match_level(self, match_level):
        # type: (MatchLevel) -> None
        self.configure.match_level = match_level

    @property
    def is_cut_provider_explicitly_set(self):
        return self._cut_provider and not (
            isinstance(self._cut_provider, NullCutProvider)
        )

    @property
    def debug_screenshot_provider(self):
        return self._debug_screenshot_provider

    @property
    def cut_provider(self):
        # type: () -> Union[FixedCutProvider, UnscaledFixedCutProvider, NullCutProvider]
        return self._cut_provider

    @cut_provider.setter
    def cut_provider(self, cutprovider):
        # type: (Union[FixedCutProvider,UnscaledFixedCutProvider,NullCutProvider])->None
        argument_guard.is_in(
            cutprovider, [FixedCutProvider, UnscaledFixedCutProvider, NullCutProvider]
        )
        self._cut_provider = cutprovider

    @property
    def _debug_screenshot_provided(self):
        # type: () -> bool
        """True if screenshots saving enabled."""
        return isinstance(self._debug_screenshot_provider, FileDebugScreenshotProvider)

    @_debug_screenshot_provided.setter
    def _debug_screenshot_provided(self, save):
        # type: (bool) -> None
        prev = self._debug_screenshot_provider
        if save:
            self._debug_screenshot_provider = FileDebugScreenshotProvider(
                prev.prefix, prev.path
            )
        else:
            self._debug_screenshot_provider = NullDebugScreenshotProvider()

    @property
    def _environment(self):
        # type: () -> AppEnvironment
        """
        Application environment is the environment (e.g., the host OS)
        which runs the application under test.

        :return: The current application environment.
        """

        app_env = AppEnvironment(
            os=self.configure.host_os,
            hosting_app=self.configure.host_app,
            display_size=self.configure.viewport_size,
            inferred=self._inferred_environment,
        )
        return app_env

    @property
    def scale_ratio(self):
        # type: () -> float
        return self._scale_provider.scale_ratio

    @scale_ratio.setter
    def scale_ratio(self, scale_ratio):
        # type: (float) -> None
        if scale_ratio:
            self._scale_provider = FixedScaleProvider(scale_ratio)
        else:
            self._scale_provider = NullScaleProvider()

    @property
    def position_provider(self):
        # type: () -> PositionProvider
        return self._position_provider

    @position_provider.setter
    def position_provider(self, provider):
        # type: (PositionProvider) -> None
        if isinstance(provider, PositionProvider):
            self._position_provider = provider
        else:
            self._position_provider = InvalidPositionProvider()

    @property
    def full_agent_id(self):
        # type: () -> Text
        """
        Gets the agent id, which identifies the current library using the SDK.

        :return: The agent id.
        """
        if self.configure.agent_id is None:
            return self.base_agent_id
        return "{0} [{1}]".format(self.configure.agent_id, self.base_agent_id)

    @property
    def agent_setup(self):
        # TODO: Implement agent_setup
        return None

    def add_property(self, name, value):
        # type: (Text, Text) -> None
        """
        Associates a key/value pair with the test. This can be used later for filtering.
        :param name: (string) The property name.
        :param value: (string) The property value
        """
        self.configure.add_property(name, value)

    def clear_properties(self):
        """
        Clears the list of custom properties.
        """
        self.configure.clear_properties()

    @property
    def is_open(self):
        # type: () -> bool
        """
        Returns whether the session is currently running.
        """
        return self._is_opened

    @staticmethod
    def log_session_results_and_raise_exception(raise_ex, results):
        logger.info("close({}): {}".format(raise_ex, results))
        results_url = results.url
        scenario_id_or_name = results.name
        app_id_or_name = results.app_name
        if results.is_unresolved:
            if results.is_new:
                logger.info(
                    "--- New test ended. \n\tPlease approve the new baseline at {}".format(
                        results_url
                    )
                )
                if raise_ex:
                    raise NewTestError(results, scenario_id_or_name, app_id_or_name)
            else:
                logger.info(
                    "--- Differences are found. \n\tSee details at {}".format(
                        results_url
                    )
                )
                if raise_ex:
                    raise DiffsFoundError(results, scenario_id_or_name, app_id_or_name)
        elif results.is_failed:
            logger.info(
                "--- Failed test ended. \n\tSee details at {}".format(results_url)
            )
            if raise_ex:
                raise TestFailedError(results, scenario_id_or_name, app_id_or_name)
        else:
            logger.info("--- Test passed. \n\tSee details at {}".format(results_url))

    def close(self, raise_ex=True):
        # type: (bool) -> Optional[TestResults]
        """
        Ends the test.

        :param raise_ex: If true, an exception will be raised for failed/new tests.
        :return: The test results.
        """
        if self.configure.is_disabled:
            logger.debug("close(): ignored (disabled)")
            return None
        try:
            logger.debug("close({})".format(raise_ex))
            if not self._is_opened:
                raise EyesError("Eyes not open")

            self._is_opened = False

            self._reset_last_screenshot()
            self._init_providers(hard_reset=True)

            # If there's no running session, we simply return the default test results.
            if not self._running_session:
                logger.debug("close(): Server session was not started")
                logger.info("close(): --- Empty test ended.")
                return TestResults(status="Failed")

            is_new_session = self._running_session.is_new_session
            results_url = self._running_session.url

            logger.info("close(): Ending server session...")
            should_save = (is_new_session and self.configure.save_new_tests) or (
                (not is_new_session) and self.configure.save_failed_tests
            )
            logger.info("close(): Automatically save session? %s" % should_save)
            results = self._server_connector.stop_session(
                self._running_session, False, should_save
            )
            results.is_new = is_new_session
            results.url = results_url
            self.log_session_results_and_raise_exception(raise_ex, results)

            return results
        finally:
            self._running_session = None
            logger.close()

    def abort(self):
        # type: () -> Optional[TestResults]
        """
        If a test is running, aborts it. Otherwise, does nothing.
        """
        if self.configure.is_disabled:
            logger.debug("abort(): ignored (disabled)")
            return
        try:
            self._reset_last_screenshot()

            if self._running_session:
                results_url = self._running_session.url

                logger.debug("abort(): Aborting session...")
                try:
                    logger.info(
                        "--- Test aborted. \n\tSee details at {}".format(results_url)
                    )
                    results = self._server_connector.stop_session(
                        self._running_session, True, False
                    )
                    results.url = results_url
                    return results
                except EyesError as e:
                    logger.info("Failed to abort server session: %s " % e)
                finally:
                    self._running_session = None
        finally:
            logger.close()

    def abort_if_not_closed(self):
        logger.deprecation("Use `abort()` instead")
        self.abort()

    def open_base(
        self,
        app_name,  # type: Text
        test_name,  # type: Text
        viewport_size=None,  # type: Optional[ViewPort]
        session_type=SessionType.SEQUENTIAL,  # type: SessionType
    ):
        # type: (...) -> None
        """
        Starts a test.

        :param app_name: The name of the application under test.
        :param test_name: The test name.
        :param viewport_size: The client's viewport size (i.e.,
                              the visible part of the document's body) or None to
                              allow any viewport size.
        :param session_type: The type of test (e.g., Progression for timing tests)
                              or Sequential by default.
        :return: An updated web driver
        :raise EyesError: If the session was already open.
        """
        logger.open_()
        if self.configure.is_disabled:
            logger.debug("open_base(): ignored (disabled)")
            return

        if self._server_connector is None:
            raise EyesError("Server connector not set.")

        # If there's no default application name, one must be provided for the current test.
        if self.configure.app_name is None:
            argument_guard.not_none(app_name)
            self.configure.app_name = app_name

        argument_guard.not_none(test_name)
        self.configure.test_name = test_name

        logger.info("\nAgent: {}\n".format(self.full_agent_id))
        logger.info(
            "open_base(%s, %s, %s, %s)"
            % (app_name, test_name, viewport_size, self.configure.failure_reports)
        )
        self.configure.session_type = session_type
        self.configure.viewport_size = viewport_size

        self._open_base()

    def _before_open(self):
        pass

    def _after_open(self):
        pass

    def _init_providers(self, hard_reset=False):
        if hard_reset:
            self._scale_provider = NullScaleProvider()
            self._position_provider = InvalidPositionProvider()
            self._cut_provider = NullCutProvider()
            self._debug_screenshot_provider = NullDebugScreenshotProvider()

        if self._scale_provider is None:
            self._scale_provider = NullScaleProvider()

        if self._position_provider is None:
            self._position_provider = InvalidPositionProvider()

        if self._cut_provider is None:
            self._cut_provider = NullCutProvider()

    def _open_base(self):
        if self.configure.is_disabled:
            logger.debug("open_base(): ignored (disabled)")
            return
        logger.open_()
        self._log_open_base()

        retry = 0
        while retry < self._MAX_ITERATIONS:
            try:
                self._validate_session_open()
                self._init_providers()

                self._is_viewport_size_set = False

                self._before_open()
                try:
                    if self.configure.viewport_size:
                        self._ensure_running_session()
                except Exception as e:
                    logger.exception(e)
                    retry += 1
                    continue

                self._is_opened = True
                self._after_open()
                return None
            except EyesError as e:
                logger.exception(e)
                logger.close()
                raise e

        raise EyesError("eyes.open_base() failed")

    def _validate_session_open(self):
        if self.is_open:
            self.abort()
            raise EyesError("A test is already running")

    def _log_open_base(self):
        logger.info("Eyes server URL is '{}'".format(self.configure.server_url))
        logger.info("Timeout = {} ms".format(self.configure._timeout))
        logger.debug("match_timeout = {} ms".format(self.configure.match_timeout))
        logger.info(
            "Default match settings = '{}' ".format(
                self.configure.default_match_settings
            )
        )
        logger.info("FailureReports = '{}' ".format(self.configure.failure_reports))

    def _create_session_start_info(self):
        # type: () -> None
        self._session_start_info = SessionStartInfo(
            agent_id=self.full_agent_id,
            session_type=self.configure.session_type,
            app_id_or_name=self.configure.app_name,
            ver_id=None,
            scenario_id_or_name=self.configure.test_name,
            batch_info=self.configure.batch,
            baseline_env_name=self.configure.baseline_env_name,
            environment_name=self.configure.environment_name,
            environment=self._environment,
            default_match_settings=self.configure.default_match_settings,
            branch_name=self.configure.branch_name,
            parent_branch_name=self.configure.parent_branch_name,
            baseline_branch_name=self.configure.baseline_branch_name,
            save_diffs=self.configure.save_diffs,
            render=self._render,
            properties=self.configure.properties,
        )

    def _start_session(self):
        # type: () -> None
        logger.debug("_start_session()")
        self.__ensure_viewport_size()

        # initialization of Eyes parameters if empty from ENV variables
        logger.info("Batch is {}".format(self.configure.batch))

        self._server_connector.update_config(
            self.get_configuration(), self.full_agent_id
        )
        self._create_session_start_info()
        # Actually start the session.
        self._running_session = self._server_connector.start_session(
            self._session_start_info
        )
        self._should_match_once_on_timeout = self._running_session.is_new_session

    def _reset_last_screenshot(self):
        # type: () -> None
        self._last_screenshot = None
        del self._user_inputs[:]

    def _ensure_running_session(self):
        if self._running_session:
            logger.debug("Session already running.")
            return

        logger.debug("No running session, calling start session...")
        self._start_session()

        output_provider = AppOutputProvider(self._get_app_output_with_screenshot)
        self._match_window_task = MatchWindowTask(
            self._server_connector,
            self._running_session,
            self.configure.match_timeout,
            eyes=self,
            app_output_provider=output_provider,
        )

    def _get_app_output_with_screenshot(self, region, last_screenshot, check_settings):
        # type: (Region, EyesScreenshot, CheckSettings) -> AppOutputWithScreenshot
        logger.info("Getting screenshot...")
        screenshot = self._get_screenshot()
        logger.info("Done getting screenshot!")
        if not region.is_size_empty:
            screenshot = screenshot.sub_screenshot(region)
            self._debug_screenshot_provider.save(screenshot.image, "sub_screenshot")

        if not self._dom_url and (
            self.configure.send_dom or check_settings.values.send_dom
        ):
            logger.info("Capturing DOM")
            dom_json = self._try_capture_dom()
            self._dom_url = self._try_post_dom_capture(dom_json)
            logger.info("Captured DOM URL: {}".format(self._dom_url))

        app_output = AppOutput(
            title=self._title, screenshot_bytes=None, dom_url=self._dom_url
        )
        result = AppOutputWithScreenshot(app_output, screenshot)
        logger.info("Done getting screenshot and DOM!")
        return result

    def _before_match_window(self):
        """
        Allow to add custom behavior after receiving response from the server
        """

    def _after_match_window(self):
        """
        Allow to add custom behavior before sending data to the server
        """

    def _check_window_base(
        self, region_provider, tag=None, ignore_mismatch=False, check_settings=None
    ):
        # type: (RegionProvider, Optional[Text], bool, CheckSettings) -> MatchResult
        if self.configure.is_disabled:
            logger.info("check_window(%s): ignored (disabled)" % tag)
            return MatchResult(as_expected=True)

        self._ensure_running_session()

        self._before_match_window()

        tag = tag if tag is not None else ""
        result = self._match_window(
            region_provider, tag, ignore_mismatch, check_settings
        )
        self._after_match_window()
        self._handle_match_result(result, tag)
        return result

    def _handle_match_result(self, result, tag):
        # type: (MatchResult, Text) -> None
        self._last_screenshot = result.screenshot
        as_expected = result.as_expected
        self._user_inputs = []
        if not as_expected:
            self._should_match_once_on_timeout = True
            if self._running_session and not self._running_session.is_new_session:
                logger.info("Window mismatch %s" % tag)
                if self.configure.failure_reports == FailureReports.IMMEDIATE:
                    raise TestFailedError(
                        "Mismatch found in '%s' of '%s'"
                        % (
                            self._session_start_info.scenario_id_or_name,
                            self._session_start_info.app_id_or_name,
                        )
                    )

    def _try_post_dom_capture(self, dom_json):
        # type: (Text) -> Optional[Text]
        """
        In case DOM data is valid uploads it to the server and return URL where it stored.
        """
        if dom_json is None:
            return None
        try:
            return self._server_connector.post_dom_capture(dom_json)
        except Exception as e:
            logger.warning(
                "Couldn't send DOM Json. Passing...\n Got next error: {}".format(e)
            )
            return None

    def _match_window(self, region_provider, tag, ignore_mismatch, check_settings):
        # type: (RegionProvider, Text, bool, CheckSettings) -> MatchResult
        # Update retry timeout if it wasn't specified.
        retry_timeout_ms = -1  # type: Num
        if check_settings:
            retry_timeout_ms = check_settings.values.timeout

        region = region_provider.get_region()
        logger.debug("params: ([{}], {}, {} ms)".format(region, tag, retry_timeout_ms))

        result = self._match_window_task.match_window(
            self._user_inputs,
            region,
            tag,
            self._should_match_once_on_timeout,
            ignore_mismatch,
            check_settings,
            retry_timeout_ms,
        )
        return result

    def __ensure_viewport_size(self):
        # type: () -> None
        """
        Assign the viewport size we need to be in the default content frame.
        """
        if self._is_viewport_size_set:
            return
        try:
            if self.configure.viewport_size is None:
                # TODO: ignore if viewport_size settled explicitly
                target_size = self._get_viewport_size()
                self.configure.viewport_size = target_size
            else:
                target_size = self.configure.viewport_size
                self._set_viewport_size(target_size)
            self._is_viewport_size_set = True
        except Exception as e:
            self._is_viewport_size_set = False
            raise_from(EyesError("Viewport has not been setup"), e)



from abc import abstractmethod
from copy import copy
from typing import Text, List, Union, Dict

import attr

from applitools.common import Region
from applitools.common.utils import argument_guard
from applitools.common.utils.compat import basestring, ABC
from applitools.common.utils.json_utils import JsonInclude
from applitools.common.validators import is_list_or_tuple

__all__ = ("VisualLocator",)


class VisualLocatorSettings(ABC):
    def __init__(self, *names):
        # type: (*Text) -> None
        self._names = list(names)
        self._first_only = True

    def first(self):
        # type: () -> VisualLocatorSettings
        clone = copy(self)
        clone._first_only = True
        return clone

    def all(self):
        # type: () -> VisualLocatorSettings
        clone = copy(self)
        clone._first_only = False
        return clone

    @property
    def is_first_only(self):
        # type: () -> bool
        return self._first_only

    @property
    def names(self):
        # type: () -> List[Text]
        return self._names


class VisualLocator(ABC):
    @staticmethod
    def name(name):
        # type: (Text) -> VisualLocatorSettings
        argument_guard.is_a(name, basestring)
        return VisualLocatorSettings(name)

    @staticmethod
    def names(*names):
        # type: (*Union[Text, List[Text]]) -> VisualLocatorSettings
        argument_guard.not_list_or_tuple(names)
        if len(names) == 1 and is_list_or_tuple(names[0]):
            names = names[0]
        argument_guard.are_(names, klass=basestring)
        return VisualLocatorSettings(*names)


LOCATORS_TYPE = Dict[Text, List[Region]]


class VisualLocatorsProvider(ABC):
    @abstractmethod
    def get_locators(self, visual_locator_settings):
        # type: (VisualLocatorSettings) -> LOCATORS_TYPE
        pass


@attr.s
class VisualLocatorsData(object):
    app_name = attr.ib(metadata={JsonInclude.THIS: True})
    image_url = attr.ib(metadata={JsonInclude.THIS: True})
    first_only = attr.ib(metadata={JsonInclude.THIS: True})
    locator_names = attr.ib(metadata={JsonInclude.THIS: True})
